@@ -66,6 +66,8 @@ data structures
 import fnmatch
 import logging
 import os
+import pkg_resources
+import re
 import sys
 import yaml
 
@@ -80,6 +82,8 @@ ITEM_FILE = 1
 ITEM_WEB = 2
 ITEM_STRING = 3
 
+ROOT_LEAF_PREFIX = "_"
+YACODIR_CACHEFILE = '.yacodir_cache'
 
 class Yaco(dict):
     """
@@ -95,7 +99,7 @@ class Yaco(dict):
 
     """
 
-    def __init__(self, data={}):
+    def __init__(self, data={}, leaf=None):
         """
         Constructor
 
@@ -105,12 +109,19 @@ class Yaco(dict):
 
         dict.__init__(self)
 
-        if isinstance(data, dict):
-            self.update(data)
-        elif isinstance(data, str) or isinstance(data, bytes):
-            self.update(yaml.load(data))
-        else:
-            raise Exception('cannot parse %s' % type(data))
+        if not data == {}:
+            to_update = None
+            if isinstance(data, dict):
+                to_update = data
+            elif isinstance(data, str) or isinstance(data, bytes):
+                to_update = yaml.load(data)
+            else:
+                raise Exception('cannot parse %s' % type(data))
+
+            if leaf is None or leaf == '':
+                self.update(to_update)
+            else:
+                self[leaf].update(to_update)
 
     def __str__(self):
         """
@@ -164,6 +175,7 @@ class Yaco(dict):
                 super(Yaco, self).__setitem__(key, value)
             else:
                 super(Yaco, self).__setitem__(key, Yaco(value))
+
         elif isinstance(value, list):
             # parse the list to see if there are dicts - which need to
             # be translated to Yaco objects
@@ -192,6 +204,36 @@ class Yaco(dict):
     def __delattr__(self, name):
         return super(Yaco, self).__delitem__(name)
 
+    def __getitem__(self, key):
+        """
+        as getattr, expect for when there is a '.' in the key.
+
+        it is possible to ask for yacoobject[''] - this is a form
+        of leaf loading - but means, give me the root. So - checking for
+        that
+        """
+        #print(  key)
+        if key == '':
+            return self
+
+        if not '.' in key:
+            return self.__getattr__(key)
+        else:
+            k1, k2 = key.split('.', 1)
+            return self.__getattr__(k1)[k2]
+
+    def __setitem__(self, key, value):
+        """
+        as setattr, except for when there is a dot in the key
+        """
+        if not '.' in key:
+            return self.__setattr__(key, value)
+        else:
+            k1, k2 = key.split('.', 1)
+            self.__getattr__(k1)[k2] = value
+
+    __delitem__ = __delattr__
+
     def simple(self):
         """
         return a simplified representation of this
@@ -310,18 +352,18 @@ class Yaco(dict):
             else:
                 super(Yaco, self).__setitem__(key, value)
 
-    __getitem__ = __getattr__
-    __setitem__ = __setattr__
-    __delitem__ = __delattr__
-
     def copy(self):
         ch = Yaco(self)
         return ch
 
-    def load(self, from_file):
+    def load(self, from_file, leaf=None):
         """
         Load this dict from_file
 
+        Note - it can load the file into a leaf, instead of the root
+        of this Yaco structure. Note - the leaf variable is a string,
+        but may contain dots (which are automatically interpreted)
+
         >>> import tempfile
         >>> tf = tempfile.NamedTemporaryFile(delete=True)
         >>> tf.close()
@@ -342,14 +384,17 @@ class Yaco(dict):
             with open(from_file, encoding='utf8') as F:
                 data = yaml.load(F)
 
-        self.update(data)
+        if leaf is None or leaf == '':
+            self.update(data)
+        else:
+            self[leaf].update(data)
 
     def pretty(self):
         """
         Return data as a pprint.pformatted string
         """
         return yaml.dump(self.get_data(), encoding='utf-8',
-                         default_flow_style=False)
+                         default_flow_style=False).rstrip()
 
     def get_data(self):
         """
@@ -385,6 +430,8 @@ class Yaco(dict):
                 continue
             if isinstance(k, (str)) and k and k[0] == '_':
                 continue
+            #print self.keys()
+            #print k, 'x' * 30
             data[k] = check_data(self[k])
         return data
 
@@ -444,17 +491,44 @@ class YacoFile(Yaco):
         super(YacoFile, self).save(self._filename)
 
 
+def _get_leaf(leaf, d, pattern):
+    """
+    Helper function to determine the leaf name
+    """
+#    print('12312341234', leaf, d, pattern)
+    xleaf = d.rsplit('/', 1)[-1].strip()
+    check_pattern = re.match('\*(\.[a-zA-Z0-9]+)$', pattern)
+    if check_pattern:
+        xten = check_pattern.groups()[0]
+        if xleaf[-len(xten):] == xten:
+            xleaf = xleaf[:-len(xten)].strip()
+    if xleaf.find(ROOT_LEAF_PREFIX) == 0:
+        return leaf
+    elif leaf.strip():
+        return '{}.{}'.format(leaf, xleaf)
+    else:
+        return xleaf
+
+
 class YacoDir(Yaco):
     """
     As Yaco, but load all files in a directory on top of each other.
 
     Order of loading is the alphanumerical sort of filenames
 
-    loading of subdirectories is depth first
+    files in subdirectories are loaded into leaves
+    e.g. a file in /tmp/test/sub/a.yaml with only (x=1) will end up as follows:
 
+        y = YacoDir('/tmp/test')
+        y.sub.x == 1
+
+
+    Note, YacoDir will try to cache itself in a .yacodir.cache file in the root
+    of the dirname if the modification date of this file is the same as the
+    directory - that will be loaded instead.
     """
 
-    def __init__(self, dirname, pattern='*.yaml'):
+    def __init__(self, dirname, pattern='*.config'):
         """
         Constructor
 
@@ -463,37 +537,113 @@ class YacoDir(Yaco):
         :param glob: a glob describing what files to load
         :type glob: string
         """
-
         dict.__init__(self)
-        self._directory = dirname
-        self._pattern = pattern
-        self.load()
+        #print("loading", dirname, pattern)
+        self.load(dirname, pattern)
 
-    def load(self):
+    def load(self, dirname, pattern):
         """
         Load from the defined directory
         """
 
-        cache = os.path.join(self._directory, '.YacoCache')
+        cachefile = os.path.join(dirname, YACODIR_CACHEFILE)
 
-        for root, dirs, files in os.walk(self._directory):
-            to_parse = sorted(fnmatch.filter(files, self._pattern))
-            lg.critical("{0} {1}".format(root, dirs))
+        # TODO: get caching to work properly :(
+        # if os.path.exists(cachefile):
+        #     if os.path.getmtime(dirname) == \
+        #             os.path.getmtime(cachefile):
+        #         #load cache
+        #         super(YacoDir, self).load(cachefile)
+        #         return
+
+        for root, dirs, files in os.walk(dirname):
+            #print('-' * 80)
+            ##import sh
+            #print(sh.ls("-l", dirname))
+            #print(root, dirs, files, pattern)
+            to_parse = sorted(fnmatch.filter(files, pattern))
+            base = root.replace(dirname, '').strip('/')
+            base = base.replace('/', '.')
+            #lg.critical("{0} {1}".format(root, dirs))
             for filename in to_parse:
                 fullname = os.path.join(root, filename)
-                lg.critical("YacoDir loading {0}".format(fullname))
-                super(YacoDir, self).load(fullname)
+                lg.debug("YacoDir loading {0}".format(fullname))
+                #print ("loadlaod", filename, fullname)
+                nleaf = _get_leaf(base, filename, pattern)
 
-        #after loading - save a cached copy!
+                with open(fullname) as F:
+                    y = yaml.load(F.read())
 
+                if nleaf == '':
+                    self.update(y)
+                else:
+                    self[nleaf].update(y)
+        #print('*' * 80)
+        #print self.pretty()
+        if self:
+            #after loading - save to cache!
+            super(YacoDir, self).save(cachefile)
 
     def save(self):
         """
         Save is disabled.
         """
-        raise Exeption("Cannot save to a YacoDir")
+        raise Exception("Cannot save to a YacoDir")
 
-class PolyYaco(object):
+
+class YacoPkg(Yaco):
+
+    def __init__(self, pkg_name, path,
+                 pattern='*.config', leaf="",
+                 base_path=None, prefix=None):
+
+
+        #lg.setLevel(logging.DEBUG)
+        lg.debug("pkg loading {} {} {}".format(pkg_name, path, pattern))
+
+        if not base_path is None:
+            if leaf:
+                leaf = leaf.strip('.') + '.'
+            leaf = path.replace(base_path, '').strip('/').replace('/', '.')
+            lg.debug("leaf: ({}) {}".format(base_path, leaf))
+
+        if not pkg_resources.resource_isdir(pkg_name, path):
+            #asssume a file:
+            lg.debug("loading file {} {}".format(pkg_name, path))
+            #print("loading file {} {}".format(pkg_name, path))
+            y = pkg_resources.resource_string(pkg_name, path)
+            self[leaf].update(yaml.load(y))
+
+        else:
+            for d in pkg_resources.resource_listdir(pkg_name, path):
+                nres = os.path.join(path, d)
+                #print('nn', nres, leaf)
+                lg.debug("checking for pkg load: {}".format(nres))
+                if pkg_resources.resource_isdir(pkg_name, nres):
+                    lg.debug("pkg load: is directory: {}".format(nres))
+                    if base_path == None:
+                        base_path = path
+                    #print('d', leaf, pkg_name, nres)
+                    y = YacoPkgDir(pkg_name, nres,
+                                   pattern=pattern,
+                                   base_path=base_path)
+                    self[leaf].update(y)
+                else:
+                    if not fnmatch.fnmatch(d, pattern):
+                        lg.debug('ignoring {}'.format(nres))
+                        continue
+                    else:
+                        lg.debug("pkg load: loading file: {}".format(nres))
+                        y =  yaml.load(pkg_resources.resource_string(pkg_name, nres))
+                        lg.debug("pkg load: got: {}".format(str(y)))
+                        this_leaf = _get_leaf(leaf, d, pattern)
+                        #print('f', leaf, path, nres, d, this_leaf)
+                        self[this_leaf].update(y)
+
+
+YacoPkgDir = YacoPkg
+
+class PolyYaco(Yaco):
     """
     A meta object that allows a composite Yaco object to be loaded
     from any number of different files which are kept as a stack of
@@ -515,150 +665,74 @@ class PolyYaco(object):
     (manually for the time being).
     """
 
-    def __init__(self, name, files=[], base=None):
+    def __init__(self, name="PY", files=[],
+                 pattern='*.config',
+                 leaf=""):
         """
 
         """
-        self._PolyYaco_base = base
-        self._PolyYaco_filenames = []
-        self._PolyYaco_yacs = []
-        self._PolyYaco_dirty = False
 
         #if not items - set a default
         if files is None:
-            self._PolyYaco_rawfiles = [
-                '/etc/{0}.yaml'.format(name),
-                '~/.config/{0}/*/*.yaml'.format(name),
-                '~/.config/{0}/*.yaml'.format(name),
-            ]
-        else:
-            self._PolyYaco_rawfiles = files
+            files = [
+                '/etc/{0}.config'.format(name),
+                '~/.config/{0}/'.format(name) ]
 
+        super(PolyYaco, self).__init__()
+        self.load(leaf, files, pattern)
 
-        self.load()
-
-    def load_file(self, fn):
-        """
-        Attemto to load the item fn as a file
-        """
-
-        if len(fn) > 504:
-            #wil not handle insanely long filenames
-            return ITEM_INVALID, None
-
-        fn = os.path.expanduser(fn)
-
-        if not os.path.exists(fn):
-            return ITEM_INVALID, None
-
-        if os.path.isdir(fn):
-            return ITEM_INVALID, None
-
-        content = Yaco()
-        content.load(fn)
-#        print 'loaded', fn, content
-        return ITEM_FILE, content
-
-    def _load(self, item):
-
-        #see if this is a local file
-        ctype, content = self.load_file(item)
-        if ctype != ITEM_INVALID:
-            return ctype, content
-
-    def load(self):
+    def load(self, leaf, files, pattern):
         """
 
         """
-        self._PolyYaco_dirty = True
-        self._PolyYaco_filenames = []
-        self._PolyYaco_yacs = []
-        self._PolyYaco_merged = None
 
-        if not self._PolyYaco_base is None:
-            self._PolyYaco_filenames.append('_base')
-            self._PolyYaco_yacs.append(Yaco(self._PolyYaco_base))
+        for filename in files:
+            filename = os.path.expanduser(filename)
 
-        for filename in self._PolyYaco_rawfiles:
-            ctype, content = self.load_file(filename)
+            y  = None
+            if filename[:6] == 'pkg://':
+                #expecting pkg://Yaco/etc/config.yaml
+                base = filename[6:]
+                pkg, loc = base.split('/', 1)
+                this_pattern=pattern
+                if '*' in loc:
+                    if '/' in loc:
+                        loc, this_pattern = loc.rsplit('/', 1)
+                    else:
+                        loc, this_pattern = '/', loc
 
-            self._PolyYaco_filenames.append(filename)
-            self._PolyYaco_yacs.append(content)
+                try:
+                    y = YacoPkg(pkg, loc, pattern=this_pattern)
+                except IOError:
+                    #file does probably not exists - ignore
+                    lg.debug("cannot load file {}".format(loc))
+                    pass
+                except ImportError:
+                    #or the complete package does not exists - one of script? ignore
+                    lg.debug("cannot find package {}".format(pkg))
 
-    def _getTop(self):
-        if len(self._PolyYaco_yacs) > 0:
-            return self._PolyYaco_filenames[-1], self._PolyYaco_yacs[-1]
-        else:
-            return None, None
+                    pass
+
+            elif os.path.isdir(filename):
+                y = YacoDir(filename, pattern = pattern)
+
+            elif os.path.isfile(filename):
+                y = Yaco()
+                y.load(filename)
+            else:
+                #nothing to load
+                continue
+
+            if not y is None:
+                self[leaf].update(y)
+        #print self.pretty()
 
     def save(self):
-        cfn, cyc = self._getTop()
-        if cfn == '_base':
-            raise Exception("Cannot save to 'base' configuration")
-        cyc.save(cfn)
-
-    def merge(self):
-        if not self._PolyYaco_dirty:
-            return self._PolyYaco_merged
-
-        y = Yaco()
-        for yy in self._PolyYaco_yacs:
-            y.update(yy)
-
-        self._PolyYaco_merged = y
-        self._PolyYaco_dirty = False
-        return y
-
-    def simple(self):
-        return self.merge().simple()
-
-    def __str__(self):
-        return str(self.merge())
-
-    def has_key(self, key):
-        return key in self.merge()
-
-    def __setattr__(self, key, value):
-        #see if this is an instance variable
-        if key[:9] == '_PolyYaco':
-            self.__dict__[key] = value
-            return
-
-        self._PolyYaco_dirty = True
-        cfn, cyc = self._getTop()
-        cyc.__setattr__(key, value)
-
-    def get(self, key, default=None):
-        try:
-            rv = self.__getattr__(key)
-        except KeyError:
-            return default
-
-        if rv == {}:
-            return default
-
-        return rv
-
-
-    def __contains__(self, key):
-        return self.merge().__contains__(key)
-
-    def __getattr__(self, key):
-        if isinstance(key, int):
-            raise Exception('No integer keys please: %s' % key)
-
-        if key[:9] == '_PolyYaco':
-            try:
-                return self.__dict__[key]
-            except KeyError:
-                raise AttributeError()
-
-        rv = self.merge().__getattr__(key)
-        return rv
-
-    __setitem__ = __setattr__
-    __getitem__ = __getattr__
-
+        lg.warning("PolyYaco save is disabled")
+        #cfn, cyc = self._getTop()
+        #if cfn == '_base':
+        #    raise Exception("Cannot save to 'base' configuration")
+        #cyc.save(cfn)
 
 
 if __name__ == "__main__":
